package backfetch

import (
	"bytes"
	"errors"
	"fmt"
	"reflect"
)

// Data data type alias
type Data = interface{}

// Node returns data,
// parameters would be prepared
type Node interface { // We don't need this
	Process() Data      // dataProcessor
	Type() reflect.Type // output type
}

// NodeFunc creates a Node from a function
type NodeFunc struct {
	name     string
	input    []Node       // Serializable
	executor interface{}  // the func
	typ      reflect.Type // just do the thing

	cached func() Data
}

// Create executor from function
//

// CreateNodeFactory creates a node constructor from function
func CreateNodeFactory(name string, fn interface{}) NodeFactory {
	fntyp := reflect.TypeOf(fn)
	// Not func? err
	return func(params ...Node) (Node, error) {
		if fntyp.NumIn() != len(params) {
			return nil, errors.New("Invalid number of params")
		}
		if fntyp.NumOut() == 0 {
			return nil, errors.New("Func it should have at least 1 output")
		}
		// Allow any type to pass
		/*for i, p := range params {
			log.Println("Matching types:", p, fntyp.In(i))
			if p.Type().Kind() == reflect.Interface {
				fntyp.In(i).Implements(p.Type())
			}
			if !p.Type().Implements(fntyp.In(i)) {
				return nil, fmt.Errorf("Type mismatch: '%s' != '%s' ", p.Type().Name(), fntyp.In(i).Name())
			}
		}*/
		// Here
		node := &NodeFunc{
			name:     name,
			input:    params,
			executor: fn,
			typ:      fntyp.Out(0), // First output
		}
		return node, nil
	}
}

//Process calls the function
func (ne *NodeFunc) Process() Data {
	if ne.cached != nil {
		return ne.cached() // cached func
	}
	// Fetch inputs first
	inList := []reflect.Value{}
	for _, in := range ne.input {
		inList = append(inList, reflect.ValueOf(in.Process()))
	}

	// :/
	out := reflect.ValueOf(ne.executor).Call(inList)

	return out[0].Interface()
}

//Type returns output type
func (ne *NodeFunc) Type() reflect.Type {
	return ne.typ
}

func (ne *NodeFunc) String() string {
	ret := bytes.NewBuffer(nil)

	fmt.Fprintf(ret, "[ %s:", ne.name)

	if len(ne.input) > 0 {
		fmt.Fprintf(ret, "Inputs: ")
	}

	for i, in := range ne.input {
		if i != 0 {
			fmt.Fprintf(ret, ", ")
		}
		fmt.Fprintf(ret, "%v", in)
	}
	fmt.Fprintf(ret, "Output: %v ]", ne.typ)
	return ret.String()
}

///////////////////////// SPECIAL CONST ////////////////////
func Const(a interface{}) *NodeFunc {
	return &NodeFunc{
		name:     "const",
		executor: func() interface{} { return a },
		typ:      reflect.TypeOf(a),
	}
}

// ConstNode basic node
/*type ConstNode struct {
	val interface{}
	typ reflect.Type
}

func Const(a interface{}) Node {
	return &ConstNode{
		val: a,
		typ: reflect.TypeOf(a),
	}
}

func (n *ConstNode) Process() Data {
	return n.val
}
func (n *ConstNode) Type() reflect.Type {
	return n.typ
}*/
