package backfetch

import (
	"errors"
	"reflect"
)

//NodeFactory type of func
// Prepares a node with specifics sources
type NodeFactory func(sources ...Node) (Node, error)

//Registry node registry
type Registry struct {
	// Still nodes
	nodeFactory map[string]NodeFactory
}

//NewRegistry creates newRegistry
func NewRegistry() *Registry {
	return &Registry{
		map[string]NodeFactory{},
	}
}

// Register a nodeFactory
func (r *Registry) Register(name string, fn interface{}) error {
	// do a precheck if it is a func
	//
	if reflect.TypeOf(fn).Kind() != reflect.Func {
		return errors.New("Should be a func?")
	}

	// It is already a factory
	if factory, ok := fn.(NodeFactory); ok {
		r.nodeFactory[name] = factory
		return nil
	}
	//Create node factory here
	r.nodeFactory[name] = CreateNodeFactory(name, fn)
	return nil
}

// Create a node with given inputs
func (r *Registry) Create(name string, params ...interface{}) (Node, error) {
	createNode, ok := r.nodeFactory[name]
	if !ok {
		return nil, errors.New("Does not exists")
	}
	// Create Node list for param
	nodeParam := []Node{}
	for _, p := range params {
		switch v := p.(type) {
		// if node executor
		case Node:
			nodeParam = append(nodeParam, v)
		default: // Check if pointer too
			nodeParam = append(nodeParam, Const(v))
			//return nil, fmt.Errorf("Unimplemented")
		}
	}
	return createNode(nodeParam...)

}
