package flow

import (
	"errors"
	"reflect"
)

// Errors
var (
	ErrNotFunc       = errors.New("Is not a function")
	ErrInvalidOutput = errors.New("Invalid output")

	globalRegistry = NewRegistry()
	Register       = globalRegistry.Register
)

type opFactory func(flow *Flow, source ...Operation) (Operation, error)

// Registry of operations
type Registry struct {
	opFactory map[string]opFactory
}

//NewRegistry creates a new Registry
func NewRegistry() *Registry {
	r := &Registry{
		map[string]opFactory{},
	}
	// Register defaults

	//r.Register("const")

	return r
}

//Register register a new operator
func (r *Registry) Register(name string, fn interface{}) error {
	fntyp := reflect.TypeOf(fn)

	if fntyp.Kind() != reflect.Func {
		return ErrNotFunc
	}
	if fntyp.NumOut() == 0 {
		return ErrInvalidOutput
	}
	// Need to have one output
	//
	// Directly add this factory?
	if f, ok := fn.(opFactory); ok {
		r.opFactory[name] = f
	}

	//Create factory from func
	r.opFactory[name] = fromFunc(name, fn)
	return nil
}

//Get a New Operation
func (r *Registry) Get(name string, flow *Flow, params ...interface{}) (Operation, error) {
	createOP, ok := r.opFactory[name]
	if !ok {
		return nil, errors.New("Does not exists")
	}
	// Create Node list for param
	opParam := make([]Operation, len(params))
	for i, p := range params {
		switch v := p.(type) {
		// if node executor
		case Operation:
			opParam[i] = v
		default: // Check if pointer too
			opParam[i] = Const(v) // Or the simple value?
		}
	}
	return createOP(flow, opParam...)

}
