package flow_test

import (
	flow "flow/x/flowt2"
	"log"
	"reflect"
	"testing"

	vecasm "github.com/gohxs/vec-benchmark/asm"
)

func TestFlow(t *testing.T) {
	f := flow.New()
	a := f.Const([]float32{1, 2, 3})
	b := f.Const([]float32{2, 2, 2})
	f.Op("mul", a, b)

	log.Println(f)
}
func TestBuild(t *testing.T) {
	var err error
	err = flow.Register("vecmul", vecmul)
	if err != nil {
		t.Fatal(err)
	}
	err = flow.Register("vecadd", vecadd)
	if err != nil {
		t.Fatal(err)
	}

	f := flow.New()

	a := f.Const([]float32{1, 2, 3})
	b := f.Const([]float32{2, 2, 2})
	m := f.Op("vecmul", []float32{1, 2, 3}, b)
	lastOp := f.Op("vecadd", a, m)

	{
		g, err := f.Build()
		if err != nil {
			t.Fatal(err)
		}
		res, err := g.Run(m, lastOp)

		test := []interface{}{
			[]float32{2, 4, 6},
			[]float32{3, 6, 9},
		}
		if !reflect.DeepEqual(test, res) {
			t.Fatal("Arrays does not match:", test, res)
		}
		t.Log("Result:", res)
	}
}
func TestImplementation(t *testing.T) {
	f := flow.New()

	in := f.Variable([]float32{2, 2, 2, 2})

	v1 := f.Const([]float32{1, 2, 3, 4})

	op1 := f.Op("vecmul", v1, v1)  // 1 2 9 16
	op2 := f.Op("vecmul", in, op1) // 2 4 18 32

	g, err := f.Build()
	if err != nil {
		t.Fatal(err)
	}
	res, err := g.Run(op1, op2)
	t.Log("Res:", res)

	test := []interface{}{
		[]float32{1, 4, 9, 16},
		[]float32{2, 8, 18, 32},
	}
	if !reflect.DeepEqual(test, res) {
		t.Fatal("Arrays does not match:", test, res)
	}
}

func TestVariable(t *testing.T) {
	f := flow.New()
	v := f.Variable([]float32{1, 2, 3})

	g, err := f.Build()
	if err != nil {
		t.Fatal(err)
	}
	res, _ := g.Run(v)
	log.Println("res", res)
	v.Set([]float32{4, 5, 6})
	res, _ = g.Run(v)
	log.Println("res", res)
}

func init() {
	flow.Register("vecmul", vecmul)
	flow.Register("vecadd", vecadd)
}

// Some funcs
func vecmul(a, b []float32) []float32 {

	sz := min(len(a), len(b))

	out := make([]float32, sz)
	vecasm.VecMulf32x8(a, b, out)
	log.Println("Mul:", a, b, "=", out)
	return out
}
func vecadd(a, b []float32) []float32 {
	sz := min(len(a), len(b))
	out := make([]float32, sz)
	for i := 0; i < sz; i++ {
		out[i] = a[i] + b[i]
	}
	return out
}

func min(p ...int) int {
	min := p[0]
	for _, v := range p[1:] {
		if min < v {
			min = v
		}
	}
	return min
}
