package flow_test

import (
	flow "flow/x/flowt2"
	"reflect"
	"testing"
)

type Stateful struct {
	i int
}

func (s *Stateful) Increment() int {
	s.i++
	return s.i
}

// Returns a function
func StatefulConstructor() func() int {
	i := 0
	return func() int {
		i++
		return i
	}
}

func TestRegistry(t *testing.T) {
	s := &Stateful{}
	r := flow.NewRegistry()
	r.Register("stateful1", s.Increment)
	r.Register("stateful2", StatefulConstructor)
	{
		f := flow.New()
		f.SetRegistry(r)
		s1op1 := f.Op("stateful1")
		s1op2 := f.Op("stateful1")

		s2op1 := f.Op("stateful2")
		s2op2 := f.Op("stateful2")

		var err error
		g, err := f.Build() // Get graph
		if err != nil {
			t.Fatal(err)
		}

		res, err := g.Run(s1op1, s1op1, s1op2) // Each time running it will increment all
		errcheck(t, err)
		t.Log("Reg:", res)
		if !reflect.DeepEqual(res, []interface{}{1, 2, 3}) {
			t.Fatal("Result mismatch")
		}

		res, err = g.Run(s2op1, s2op1, s2op2) // Each time it will increment instance
		errcheck(t, err)
		t.Log("Reg:", res)
		if !reflect.DeepEqual(res, []interface{}{1, 2, 1}) {
			t.Fatal("Result mismatch")
		}
	}
}

func errcheck(t *testing.T, err error) {
	if err != nil {
		t.Fatal(err)
	}
}
